#define F_CPU 1000000UL

#include <stddef.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include <util/delay.h>

// Pulse length in timer1 units
volatile uint16_t tPulse;

// Convert ?s to timer1 ticks
// For F_CPU=8MHz and /8 prescaler no conversion is needed
#define US2TIMER1(x) (x)

#define REG
#ifdef REG
// ser-par reg
#define OUT_NUM 2
//#define IN_NUM 2
//#define REG_LOAD	PB0
#define REG_LATCH	PB5
#define REG_CLK		PB3
#define REG_DATA	PB4
#define REG_IN(bit)  DDRB  &= ~(1<<bit)
#define REG_OUT(bit) DDRB  |=  (1<<bit)
#define REG_GET(bit) PINB  &   (1<<bit)
#define REG_SET(bit) PORTB |=  (1<<bit)
#define REG_CLR(bit) PORTB &= ~(1<<bit)

#ifdef REG_LOAD
uint8_t reg_recvbyte()
{
	uint8_t mask = 0x80;
	uint8_t data = 0;
	do {
		// read bit (D7 is already at output)
		if (REG_GET(REG_DATA))
			data |= mask;
		// pulse clock
		REG_SET(REG_CLK);
		_delay_us(0.2);
		REG_CLR(REG_CLK);
		// next bit
		mask >>= 1;
	} while (mask);
	return data;
}

void reg_recvdata(uint8_t *data, uint8_t count)
{
	uint8_t i;

	// set data pin to input
	REG_IN(REG_DATA);
	// load input to register
	REG_SET(REG_LOAD);
	_delay_us(0.2);
	// receive data
	for (i = 0; i < count; i++)
		data[i] = reg_recvbyte();
	REG_CLR(REG_LOAD);
}
#endif

#ifdef REG_LATCH
void reg_sendbyte(uint8_t data)
{
	uint8_t mask = 0x80;
	do {
		// set data pin
		if (data & mask)
			REG_SET(REG_DATA);
		else
			REG_CLR(REG_DATA);
		// pulse clock
		REG_SET(REG_CLK);
		_delay_us(0.2);
		REG_CLR(REG_CLK);
		// next bit
		mask >>= 1;
	} while (mask);
	// clear data pin
	REG_CLR(REG_DATA);
}

void reg_senddata(uint8_t *data, uint8_t count)
{
	uint8_t i;

	// set data pin to output
	REG_OUT(REG_DATA);
	// send data
	for (i = count; i > 0; i--)
		reg_sendbyte(data?data[i-1]:0);
	// propagate data to output
	REG_SET(REG_LATCH);
	_delay_us(0.2);
	REG_CLR(REG_LATCH);
}
#endif

void reg_init()
{
	// clock is output
	REG_OUT(REG_CLK);
	// clear clock
	REG_CLR(REG_CLK);
#ifdef REG_LOAD
	// load is output
	REG_OUT(REG_LOAD);
	// clear load
	REG_CLR(REG_LOAD);
#endif
#ifdef REG_LATCH
	// latch is output
	REG_OUT(REG_LATCH);
	// clear latch
	REG_CLR(REG_LATCH);
	reg_senddata(NULL, OUT_NUM);
#endif
}
#endif

uint16_t d2fx(uint16_t dec)
{
	uint16_t fx = 0;
	uint8_t i = 0;
	do {
		fx |= (dec % 10) << i;
		i += 4;
	} while ((dec /= 10) > 0);
	return fx;
}

int main(void)
{
	/*DDRB = 0xFF; // Debugging leds
	PORTB = 0xFF;// Leds off*/

#ifdef REG
	// register init
	reg_init();
#endif

	// Timer0 16-bit width, Input Capture enabled, IC rising edge select
	TCCR0A |= (1<<TCW0)|(1<<ICEN0)|(1<<ICES0);
	// clear interrupt-flag
	TIFR |= (1 << ICF0);
	// enable Timer0 input capture interrupt
	TIMSK |= (1 << TICIE0);
	// Timer0 start @CK
	TCCR0B |= (1<<CS00);

	sleep_enable();
	sei();
	while(1) {
		sleep_cpu();
	}
}

ISR(TIMER0_CAPT_vect)
{
	static int16_t tStart;
	
	// read ICR0
	cli();
	uint16_t t = (uint16_t)(OCR0B << 8) | OCR0A;
	sei();
	if (TCCR0A & (1<<ICES0)) {
		// falling edge next
		TCCR0A &= ~(1<<ICES0);
		tStart = t;
	} else {
		// rising edge next
		TCCR0A |= (1<<ICES0);
		tPulse = t - tStart;
		if (tPulse >= US2TIMER1(750) && tPulse <= US2TIMER1(2250)) {
			// Do something...
			// led bar
			// Here: the pulse time is scaled and output on PORTB leds
			/*int ledBit = (((int)t - US2TIMER1(1000)) * 8) / US2TIMER1(1000);
			if (ledBit < 1) {
				ledBit = 1;
			}
			if (ledBit > 8) {
				ledBit = 8;
			}
			PORTB = ~(uint8_t)((1<<ledBit) - 1); // Active low leds
			*/
			// brake reverse leds
			/*if (t <= 1400) // reverse
				PORTB = ~(1<<PB1);
			else if (t >= 1600) // forward
				PORTB = 0xFF;
			else // brake
				PORTB = ~(1<<PB0);
			*/
			// 4 pos 7-seg display
			// convert to fake hex
			tPulse--;
			tPulse = d2fx(tPulse);
#ifdef REG
			reg_senddata((uint8_t*)&tPulse, 2);
#endif
		}
	}
}
