/*
 * RC_control.c
 *
 * Created: 2014-08-15 14:26:39
 *  Author: nnobody
 */ 

// Int RC/Ext Crystal 8MHz, CKDIV8
#define F_CPU 1000000UL

#include <stddef.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include <util/delay.h>
#include <avr/wdt.h>

#define REM_MODE	/* remember last mode set */

#define MODE_MAN 0
#define MODE_MID 1
#define MODE_AUT 2
#define MODE_MAX MODE_AUT
#ifdef REM_MODE
#include <avr/eeprom.h>
uint8_t EEMEM eemode = MODE_MAN;
#endif
volatile uint8_t mode = MODE_MAN;

#define STROKE_MIN  990
#define STROKE_MAX 2008
#define STROKE_INC   10

#define BTN (1<<PD2)
#define LED_MAN PB3
#define LED_MID PB4
#define LED_AUT PB5
#define LEDS ((1<<LED_MAN)|(1<<LED_MID)|(1<<LED_AUT))

volatile uint8_t bdelay = 0;
volatile uint8_t timer = 0;
int8_t inc;

void changemode()
{
	if (mode > MODE_MAX)
		mode = 0;
#ifdef REM_MODE
	eeprom_update_byte(&eemode, mode);
#endif

	// mode LED
	PORTB = (PORTB | LEDS) & ~(1<<(mode+LED_MAN));

	if (mode == MODE_MID)
		OCR1A = 1500;
	else if (mode == MODE_AUT)
		inc = STROKE_INC;
}

void timer_start()
{
	// Timer1 TOP value (50Hz @ 1MHz)
	ICR1 = 20000;
	// Timer1 initial compare value
	//OCR1A = STROKE_MIN;
	// Timer1 clear OC1A on match/set on bottom, Fast PWM to ICR1
	TCCR1A |= (1<<COM1A1)|(1<<WGM11);
	TCCR1B |= (1<<WGM12)|(1<<WGM13);
	// Timer1 enable compare A match INT
	TIMSK |= (1<<OCIE1A);
	// Timer1 start @CK
	TCCR1B |= (1<<CS10);
	// Timer1 start @CK/8
	//TCCR1B |= (1<<CS11);
	timer = 1;
}

int main(void)
{
	/* disable watchdog */
	MCUSR = 0;
	wdt_disable();

	// PB1 is Timer1 PWM out
	DDRB |= (1<<PB1);
	// LEDs output
	DDRB |= LEDS;
	// button pull-up
	PORTD |= BTN;

	// ADC left adj, ADC ref is AVCC, ADC0 chan
	ADMUX |= (1<<ADLAR)|(1<<REFS0);
	// ADC auto trig on Timer1 overflow (broken when used with CPU sleep)
	//ADCSRB |= (1<<ADTS1)|(1<<ADTS2);

#ifdef REM_MODE
	mode = eeprom_read_byte(&eemode);
#endif
	changemode();

	if (mode == MODE_MID)				// value is well known, timer may be started
		timer_start();
	else								// otherwise initial value must be determined first
		// ADC enable, ADC INT enable, reset INT flag, ADC start conversion, @CK/16
		ADCSRA = (1<<ADEN)|(1<<ADIE)|(1<<ADIF)|(1<<ADSC)|(1<<ADPS2);

	/* button interrupt */
	MCUCR |= (1<<ISC01);	// falling edge
	GICR |= (1<<INT0);		// INT0 is button interrupt

	sleep_enable();
	sei();
	while(1) {
		sleep_cpu();
	}
}

ISR(TIMER1_COMPA_vect)
{
	cli();
	
	if (bdelay > 0) {
		if (bdelay > 1) { // debounce 15-30ms
			if (!(PINB & BTN)) { // still pressed
				mode++;
				changemode();
			}
			bdelay = 0;
		} else
			bdelay++;
	}

	switch (mode) {
		case MODE_MAN: // manual control
			// Enable ADC and start conversion
			ADCSRA = (1<<ADEN)|(1<<ADIE)|(1<<ADIF)|(1<<ADSC)|(1<<ADPS2);
			break;
		case MODE_AUT: { // auto full stroke
			OCR1A += inc;
			if (OCR1A > STROKE_MAX) {
				OCR1A = STROKE_MAX;
				inc *= -1;
			} else if (OCR1A < STROKE_MIN) {
				OCR1A = STROKE_MIN;
				inc *= -1;
			}
		    } break;
		default:
			break;
	}

	sei();
}

ISR(ADC_vect)
{
	cli();

	// use 8MSB's only
	OCR1A = (ADCH << 2) + STROKE_MIN;

	// Disable ADC to prevent idle mode auto triggering (h/w bug)
	ADCSRA = 0;
	// Start Timer1 if not started yet
	if (!timer)
		timer_start();

	sei();
}

ISR(INT0_vect)
{
	if (bdelay == 0)
		bdelay = 1;
}
